from collections import namedtuple
from typing import List, Tuple, Union, Dict, Optional
from .gtftools import GtfWriter
from .gfftools import GffWriter
from .fasta import FastaWriter
from .dataclass import GenericFeature, FeatureArray, Chromosome
from .feature_conversion import generic_to_gtf_feature, generic_to_gff_feature


GenbankText = namedtuple('GenbankText', 'locus_text features_text origin_text')


def get_seqname(locus_text: str) -> str:
    """
    Args:
        locus_text:
            The LOCUS (i.e. header section) of genbank file

    Returns:
        The seqname of chromosome
    """
    line1 = locus_text[:locus_text.find('\n')]
    return line1[len('LOCUS'):].lstrip().split(' '*3)[0]


def get_circular(locus_text: str) -> bool:
    """
    Args:
        locus_text:
            The LOCUS (i.e. header section) of genbank file
    """
    line1 = locus_text[:locus_text.find('\n')]
    if ' circular ' in line1: return True
    return False


def split_features_text(features_text: str) -> List[str]:
    """
    Args:
        features_text:
            The FEATURES (i.e. annotation) section of genbank file, for example:

            'FEATURES             Location/Qualifiers
            '     source          1..168903
            '                     /organism="Enterobacteria phage T4"
            '                     /mol_type="genomic DNA"
            '                     /db_xref="taxon:10665"
            '     CDS             complement(12..2189)
            '                     /gene="rIIA"
            '                     /locus_tag="T4p001"
            '                     /db_xref="GeneID:1258593"

    Returns:
        Break it into a list of single feature text, for example

        feature_text_1:
            '     source          1..168903
            '                     /organism="Enterobacteria phage T4"
            '                     /mol_type="genomic DNA"
            '                     /db_xref="taxon:10665"

        feature_text_2:
            '     CDS             complement(12..2189)
            '                     /gene="rIIA"
            '                     /locus_tag="T4p001"
            '                     /db_xref="GeneID:1258593"
    """
    list_ = []
    for line in features_text.splitlines()[1:]:  # first line is useless
        if line.strip() == '':  # skip empty line
            continue
        if line.startswith(' '*21):
            list_[-1] += line + '\n'
        else:
            list_.append(line + '\n')
    return list_


def get_feature_type(feature_text: str) -> str:
    """
    Args:
        feature_text: endswith '\n'
            For example:
            '     CDS             complement(12..2189)
            '                     /gene="rIIA"
            '                     /locus_tag="T4p001"
            '                     /db_xref="GeneID:1258593"

    Returns:
        The example would be 'CDS'
    """
    line1 = feature_text.splitlines()[0]
    return line1.strip().split()[0]


def get_location_string(feature_text: str) -> str:
    """
    Args:
        feature_text: endswith '\n'
            For example:
            '     CDS             complement(join(<360626..360849,360919..360948,
            '                     361067..361220,361292..361470,361523..>361555))
            '                     /gene="rIIA"
            '                     /locus_tag="T4p001"
            '                     /db_xref="GeneID:1258593"

    Returns:
        The example would be
        'complement(join(<360626..360849,360919..360948,361067..361220,361292..361470,361523..>361555))'
    """
    # The location string is before ' '*21 + '/', sometimes could be multiple lines
    pos = feature_text.find(' ' * 21 + '/')
    ret = feature_text[:pos]

    # Multiple -> single line
    ret = ret.replace('\n' + ' ' * 21, '')

    # Remove the feature type such as 'CDS'
    ret = ret.strip().split()[1]

    return ret


def get_feature_location(feature_text: str) -> Tuple[int, int, str, List, bool, bool]:
    """
    Args:
        feature_text: endswith '\n'
            For example:
            '     CDS             complement(join(<360626..360849,360919..360948,
            '                     361067..361220,361292..361470,361523..>361555))
            '                     /gene="rIIA"
            '                     /locus_tag="T4p001"
            '                     /db_xref="GeneID:1258593"

    Returns:
        start: int
            The example would be 12

        end: int
            The example would be 2189

        strand: str, '+' or '-'
            The example would be '-'

        regions: list of tuple (int, int, str)
            Indicating start, end, strand of each region (intron)
            The example would be [(12, 2189, '-')]

        partial_start: bool

        partial_end: bool
    """
    locstr = get_location_string(feature_text)

    if locstr.startswith('complement('):
        # Remove 'complement(' and ')'
        locstr = locstr[len('complement('):-1]
        all_complement = True
    else:
        all_complement = False

    if locstr.startswith('join('):
        # Remove 'join(' and ')'
        locstr = locstr[len('join('):-1]

    # loclist = list of strings
    # e.g. ["complement(2853..2990)", "complement(2458..2802))"]
    loclist = locstr.split(',')

    partial_start, partial_end = False, False
    regions = []  # e.g. [(100, 200, '-'), (<300, 400, '+'), (500, >600, '+')]
    for i, s in enumerate(loclist):
        # Tell the strand
        if s.startswith('complement('):
            # Remove 'complement(' and ')'
            s = s[len('complement('):-1]
            c = '-'           # c = strand
        elif all_complement:
            c = '-'
        else:
            c = '+'

        a, b = s.split('..') if ('..' in s) else (s, s)  # a is start, b is end

        # First start has '<' --> partial start
        if i == 0 and a.startswith('<'):
            a = a[1:]
            partial_start = True

        # Last end has '>' --> partial end
        if i == len(loclist) - 1 and b.startswith('>'):
            b = b[1:]
            partial_end = True

        a, b = int(a), int(b)

        if a > b:
            a, b = b, a  # a must be < b

        regions.append((a, b, c))

    regions = sorted(regions, key=lambda x: x[0])
    start, end, strand = regions[0][0], regions[-1][1], regions[0][2]

    return start, end, strand, regions, partial_start, partial_end


def get_feature_attributes(feature_text: str) -> List[Tuple[str, str]]:
    """
    Args:
        feature_text: endswith '\n'
            For example:
            '     CDS             complement(12..2189)
            '                     /gene="rIIA"
            '                     /locus_tag="T4p001"
            '                     /db_xref="GeneID:1258593"

    Returns: list of tuples of (key, value) pairs
        The example would be
        [
            ('gene', 'rIIA'),
            ('locus_tag', 'T4p001'),
            ('db_xref', 'GeneID:1258593')
        ]
    """
    # If no attribute at all, return empty list
    if (' '*21 + '/') not in feature_text:
        return []

    # Remove the location information before ' '*21 + '/'
    pos = feature_text.find(' '*21 + '/')
    feature_text = feature_text[pos:]

    # Parse the feature_text into a list of str, where each str is 'key="text"' or 'key=value'
    attr_list = list()
    for line in feature_text.splitlines():
        if line.startswith(' '*21+'/') and '=' in line:
            attr_list.append(line[22:])
        else:
            # Add a blank space ' ' for changing line
            attr_list[-1] += ' ' + line.lstrip()

    # Unpack key="value" or key=value into a list of tuples (key, value)
    for i, attr in enumerate(attr_list):

        # With quote, value is text
        if '="' in attr:
            pos = attr.find('="')
            key = attr[:pos]
            val = attr[pos+2:-1]

        # Without quote
        elif '=' in attr:
            key, val = attr.split('=')
            if val.isdigit():
                val = int(val)
        else:
            continue

        # Protein sequence should not contain blank space
        if key == 'translation':
            val = val.replace(' ', '')

        attr_list[i] = (key, val)

    return attr_list


def construct_generic_feature(feature_text: str, seqname: str) -> GenericFeature:
    """
    Construct a GenericFeature object from a single feature text
    Introns (joined regions) are supported

    Args:
        feature_text: endswith '\n'
            For example:
            '     CDS             complement(12..2189)
            '                     /gene="rIIA"
            '                     /locus_tag="T4p001"
            '                     /db_xref="GeneID:1258593"

        seqname: str
    """

    start, end, strand, regions, partial_start, partial_end = get_feature_location(feature_text)

    return GenericFeature(
        seqname=seqname,
        type_=get_feature_type(feature_text),
        start=start,
        end=end,
        strand=strand,
        regions=regions,
        frame=1,
        attributes=get_feature_attributes(feature_text),
        partial_start=partial_start,
        partial_end=partial_end
    )


def contruct_feature_array(
        features_text: str, seqname: str, genome_size: int, circular: bool) -> FeatureArray:
    """
    Construct a FeatureArray from the complete FEATURES section of genbank file

    Args:
        features_text: str
            The FEATURES (i.e. annotation) section of genbank file

            For example:
            'FEATURES             Location/Qualifiers
            '     source          1..168903
            '                     /organism="Enterobacteria phage T4"
            '                     /mol_type="genomic DNA"
            '                     /db_xref="taxon:10665"
            '     CDS             complement(12..2189)
            '                     /gene="rIIA"
            '                     /locus_tag="T4p001"
            '                     /db_xref="GeneID:1258593"

        seqname: str

        genome_size: int

        circular: bool
    """
    features = []

    for feature_text in split_features_text(features_text):

        features.append(
            construct_generic_feature(feature_text, seqname)
        )

    return FeatureArray(
        seqname=seqname, genome_size=genome_size,
        features=features, circular=circular)


def get_sequence(origin_text: str) -> str:
    """
    Args:
        origin_text: str
            The ORIGIN (i.e. sequence section) of genbank file

    Returns: str
        The genome sequence
    """
    lines = []

    for line in origin_text.splitlines():
        lines.append(line[9:].replace(' ', ''))

    return ''.join(lines)


def is_valid_first_line_of_feature(line: str) -> bool:

    if len(line) <= 21:
        return False

    if ' ' in line[:21].strip():
        return False

    if not line[21].isdigit() and \
            not line[21:].startswith('complement') and \
            not line[21:].startswith('join') and \
            not line[21] == '<':
        return False

    return True


def get_valid_features_text(genbank_text: GenbankText) -> str:

    lines = []

    for line in genbank_text.features_text.splitlines():

        if line == 'FEATURES             Location/Qualifiers':
            lines.append(line)

        elif line.startswith(' ' * 21):
            lines.append(line)

        elif line.startswith(' ' * 5):
            if is_valid_first_line_of_feature(line=line):
                lines.append(line)

    return '\n'.join(lines)


def construct_chromosome(genbank_text: GenbankText) -> Chromosome:

    locus_text = genbank_text.locus_text
    features_text = get_valid_features_text(genbank_text=genbank_text)
    origin_text = genbank_text.origin_text

    seqname = get_seqname(locus_text=locus_text)
    sequence = get_sequence(origin_text=origin_text)
    circular = get_circular(locus_text=locus_text)

    features = contruct_feature_array(
        features_text=features_text,
        seqname=seqname,
        genome_size=len(sequence),
        circular=circular)

    return Chromosome(
        seqname=seqname,
        sequence=sequence,
        features=features,
        circular=circular,
        genbank_locus_text=locus_text)


class GenbankTextParser:

    def __init__(self, file: str):
        """
        Args:
            file: path-like
        """
        self.__gbk = open(file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def __iter__(self):
        self.reset()
        return self

    def __next__(self):
        r = self.next()
        if r:
            return r
        else:  # r is None
            raise StopIteration

    def reset(self):
        self.__gbk.seek(0)

    def next(self) -> Optional[GenbankText]:
        pos = self.__gbk.tell()
        if not self.__gbk.readline().startswith('LOCUS'):
            return None

        self.__gbk.seek(pos)
        lines = []
        while True:
            line = self.__gbk.readline()
            if line.startswith('//'):
                break
            line = line[:-1]  # remove '\n'
            lines.append(line)
        text = '\n'.join(lines)

        locus_text = text[0:text.find('FEATURES')]
        features_text = text[text.find('FEATURES'):text.find('ORIGIN')]
        origin_text = text[text.find('ORIGIN'):]

        return GenbankText(locus_text, features_text, origin_text)

    def close(self):
        self.__gbk.close()


class GenbankParser:

    def __init__(self, file):

        self.__text_parser = GenbankTextParser(file=file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def __iter__(self):
        self.reset()
        return self

    def __next__(self):
        r = self.next()
        if r:
            return r
        else:  # r is None
            raise StopIteration

    def reset(self):
        self.__text_parser.reset()

    def next(self) -> Optional[Chromosome]:
        genbank_text = self.__text_parser.next()
        if genbank_text is None:
            return None
        else:
            return construct_chromosome(genbank_text)

    def close(self):
        self.__text_parser.close()


def read_genbank(
        file: str,
        as_dict: bool = False) -> \
        Union[List[Chromosome], Dict[str, Chromosome]]:
    """
    Read a genbank file into Chromosome ojbects, i.e. annotated genome

    Args:
        file: path-like
            The input genbank file

        as_dict:
            Return as dictionary instead of list

    Returns: list of Chromosome objects, or dict

        [Chromosome_1, Chromosome_2, ...]

        or

        {
            Chromosome_1.seqname: Chromosome_1,
            Chromosome_2.seqname: Chromosome_2, ...
        }
    """
    with GenbankParser(file) as parser:
        chromosomes = [chrom for chrom in parser]

    if as_dict:
        chromosomes = {chrom.seqname: chrom for chrom in chromosomes}

    return chromosomes


def genbank_to_fasta(file: str, output: str):
    """
    Args:
        file: path-like
            Input genbank file

        output: path-like
            Output fna file
    """
    with FastaWriter(output) as writer:
        with GenbankParser(file) as parser:
            for chromosome in parser:
                writer.write(chromosome.seqname, chromosome.sequence)


DEFAULT_SKIP_TYPES = ['source']
DEFAULT_SKIP_ATTRIBUTES = ['translation', 'codon_start', 'transl_table']


def genbank_to_gtf(
        file: str, output: str,
        skip_types: Optional[Union[str, List[str]]] = None,
        skip_attributes: Optional[Union[str, List[str]]] = None):
    """
    Extract features in the genbank file and write them into a GTF file

    Args:
        file: path-like
            Input genbank file

        output: path-like
            Output GTF file

        skip_types:
            Feature of types not to be included

        skip_attributes:
            Attributes not to be included
    """
    if skip_types is None:
        skip_types = DEFAULT_SKIP_TYPES
    elif type(skip_types) is str:
        skip_types = [skip_types]

    if skip_attributes is None:
        skip_attributes = DEFAULT_SKIP_ATTRIBUTES
    elif type(skip_attributes) is str:
        skip_attributes = [skip_attributes]

    with GtfWriter(output) as writer:
        with GenbankParser(file) as parser:
            for chromosome in parser:
                for feature in chromosome.features:
                    if feature.type in skip_types:
                        continue
                    for key in skip_attributes:
                        feature.remove_attribute(key)
                    gtf_feature = generic_to_gtf_feature(feature)
                    writer.write(gtf_feature)


def genbank_to_gff(
        file: str, output: str,
        skip_types: Optional[Union[str, List[str]]] = None,
        skip_attributes: Optional[Union[str, List[str]]] = None):
    """
    Extract features in the genbank file and write them into a GFF3 file

    Args:
        file: path-like
            Input genbank file

        output: path-like
            Output GFF file

        skip_types:
            Feature of types not to be included

        skip_attributes:
            Attributes not to be included
    """
    if skip_types is None:
        skip_types = DEFAULT_SKIP_TYPES
    elif type(skip_types) is str:
        skip_types = [skip_types]

    if skip_attributes is None:
        skip_attributes = DEFAULT_SKIP_ATTRIBUTES
    elif type(skip_attributes) is str:
        skip_attributes = [skip_attributes]

    with GffWriter(output) as writer:
        with GenbankParser(file) as parser:
            for chromosome in parser:
                for feature in chromosome.features:
                    if feature.type in skip_types:
                        continue
                    for key in skip_attributes:
                        feature.remove_attribute(key)
                    gff_feature = generic_to_gff_feature(feature)
                    writer.write(gff_feature)



import os
import subprocess
from .lowlevel import gzip, printf


def count_reads(file: str, mapped: bool = True) -> int:
    """
    Count the reads in a file

    Supported format: fasta, fastq and sam/bam, compressed gz is also supported

    Args:
        file: path-like

        mapped: bool
            If True, only count reads that are mapped in a sam or bam file
            If False, count all reads

    Returns:
        The number of reads or sequences contained in the input file
    """
    is_gz = False
    if file.endswith('.gz'):
        gzip(file, keep=True)
        file = file[:-3]
        is_gz = True

    ftypes = ['fq', 'fastq', 'fa', 'fasta', 'sam', 'bam']
    assert file.split('.')[-1] in ftypes, 'Invalid input file extention.'

    with open(file, 'r') as fh:
        # FASTQ: Use number of lines to count the reads in fastq file
        if file.endswith('.fq') or file.endswith('.fastq'):
            i = 0
            while fh.readline() != '':
                i += 1
            if i % 4 != 0:
                printf('Warning: Number of lines in the fastq file is not multiple of 4.')
            count = int(i / 4)

        # FASTA: Use '>' to count the reads in fasta file
        elif file.endswith('.fa') or file.endswith('.fasta'):
            i = 0
            while True:
                line = fh.readline()
                if line == '': break
                if line.startswith('>'):
                    i += 1
            count = i

        # SAM/BAM: Use 'samtools view -c' command to count reads
        else:
            mapped = ['', '-F 4 '][mapped]
            cmd = f"samtools view -c {mapped}{file}"
            count = int(subprocess.check_output(cmd, shell=True))

    if is_gz:
        os.remove(file)

    return count




import os
from typing import Optional, Union, List, Dict
from .gtftools import GtfWriter
from .lowlevel import call, _temp, printf
from .fasta import FastaParser, FastaWriter


def bedtools_multicov(
        bed: str,
        bams: Union[str, List[str]],
        output: str):
    """
    Wrapper function of the command "bedtools multicov -bams <bams> -bed <bed> > <output>"
    Adds a header line to the <output> file

    Args:
        bed: path-like
            The bed file, or any other interval file accepted by the bedtools

        bams: path-like or list of paths
            The bam file, or any other mapped read files accepted by the bedtools

        output: path-like
            The output tab-separated file (tsv) with headers according to the input file type
            Currently supports the header of the following file types:
                bed file: chrom \t start \t end
                gtf file: seqname \t source \t feature \t start \t end \t score \t strand \t frame \t attribute
    """
    if isinstance(bams, list):
        bams = ' '.join(bams)

    call(f"bedtools multicov -bams {bams} -bed {bed} > {output}")

    # Read the data of the bedtools output
    with open(output, 'r') as fh:
        data = fh.read()
    call(f"rm {output}")

    # Write the header line to the output file
    with open(output, 'w') as fh:
        # Replace ' ' with '\t'
        bams = '\t'.join(bams.split(' '))
        if bed.endswith('.bed'):
            fh.write(f"chrom\tstart\tend\t{bams}\n")
        if bed.endswith('.gtf'):
            fh.write(f"seqname\tsource\tfeature\tstart\tend\tscore\tstrand\tframe\tattribute\t{bams}\n")
        fh.write(data)


def bowtie2_mapping(
        ref: str,
        fq1: str,
        sam: str,
        fq2: Optional[str] = None):
    """
    Wrapper function of bowtie2 mapping

    Args:
        ref: path-like
            The reference fasta

        fq1: path-like
            The read-1 fastq

        fq2: path-like
            The read-2 fastq. If none, use <fq1> for unpaired mapping

        sam: path-like
            The output SAM file
    """
    # Build the .bt2 index files
    call(f"bowtie2-build {ref} ref > {ref}_bowtie2_build.log")

    log = sam[:-4]+'.log'
    if fq2 is None:
        # Unpaired mapping
        cmd = f"bowtie2 -x ref -U {fq1} -S {sam} 2> {log}"
    else:
        # Paired-end mapping
        cmd = f"bowtie2 -x ref -1 {fq1} -2 {fq2} -S {sam} 2> {log}"
    call(cmd)

    # Remove the .bt2 index files
    call('rm *.bt2')


def bwa_mapping(
        ref: str,
        fq1: str,
        sam: str,
        fq2: Optional[str] = None,
        threads: int = 4,
        score: int = 30):
    """
    Wrapper function of BWA mapping

    Args:
        ref: path-like
            The reference fasta

        fq1: path-like
            The read-1 fastq

        fq2: path-like
            The read-2 fastq. If none, use <fq1> for unpaired mapping

        sam: path-like
            The output SAM file

        threads:
            Number of CPUs

        score:
            Don’t output alignment with score lower than <score>
    """
    # Build index
    call(f"bwa index {ref}")

    if fq2 is None:
        # Unpaired
        cmd = f"bwa mem -t {threads} -T {score} {ref} {fq1} > {sam}"
    else:
        # Paired-end
        cmd = f"bwa mem -t {threads} -T {score} {ref} {fq1} {fq2} > {sam}"
    call(cmd)

    # Remove the index files
    call('rm ref.*')


def metaspades(
        fq1: str,
        fq2: str,
        output: str,
        min_contig_length: int = 1000,
        threads: int = 16,
        memory: int = 250):
    """
    Wrapper function of the command "spades.py --meta".

    Args:
        fq1: path-like
            The read-1 fastq file

        fq2: path-like
            The read-2 fastq file

        output: path-like
            The output directory of metaspades
            The output fasta file containing assembled contigs

        min_contig_length:
            Minimum (inclusive) contig length in bp

        threads:
            Number of CPUs

        memory:
            Number of Gb of RAM
    """
    call(f"spades.py --meta -1 {fq1} -2 {fq2} -o {output} --threads {threads} --memory {memory} > {output}.log")

    printf(f"Retrieve contigs (>= {min_contig_length} bp) from {output}/contigs.fasta -> {output}.fa")

    with FastaParser(f"{output}/contigs.fasta") as parser:
        with FastaWriter(f"{output}.fa") as writer:
            for head, seq in parser:
                if len(seq) >= min_contig_length:
                    writer.write(head, seq)


def _rename_contig_id(
        genbank: str,
        contig_dict: Dict[str, str]):
    """
    Args:
        genbank: path-like

        contig_dict:
            {'contig_1_52671bp': 'assembly=control_contigs;id=NODE_400_length_52671_cov_411.055515;len=52671'}
               (short header)        (original header)
    """
    temp_gb = _temp('temp', '.gb')  # for example, temp000.gb
    with open(genbank) as reader:
        with open(temp_gb, 'w') as writer:
            for line in reader:
                if line.startswith('LOCUS       '):
                    key = line.split()[1]
                    writer.write(
                        line.replace(key, contig_dict[key])
                    )
                else:
                    writer.write(line)
    os.remove(genbank)
    os.rename(temp_gb, genbank)


def prokka(
        fasta: str,
        outdir: str,
        kingdom: str,
        locus_tag: str = 'LOCUS',
        proteins: str = '',
        metagenome: bool = True,
        evalue: float = 1e-6,
        threads: int = 4,
        log: Optional[str] = None,
        keep: bool = True):
    """
    A wrapper function for Prokka 1.12 (rapid bacterial genome annotation)

    This wrapper function takes a fasta and outputs a genbank file with the name <outdir>.gb

    Args
        fasta: path-like
            The input fasta to be annotated

        outdir: path-like
            The output directory and the genbank file name

        kingdom:
            'Archaea', 'Bacteria', 'Mitochondria' or 'Viruses'

        locus_tag

        proteins: path-like
            FASTA or GBK file to use as 1st priority

        metagenome:
            Improve gene predictions for highly fragmented genomes

        evalue:
            Similarity e-value cut-off

        threads:
            Number of CPUs to be used (0 = all)

        log: path-like
            The log file for stderr

        keep:
            Whether or not to keep all data in the <outdir>
    """
    # Create a temp fasta file to shorten the header because prokka does not take long headers
    # Also create a dictionary to index the shortened header
    temp_fa = _temp('temp', '.fa')  # for example, temp000.fa
    contig_dict = dict()
    with FastaParser(fasta) as parser:
        with FastaWriter(temp_fa) as writer:
            i = 0
            for head, seq in parser:
                new_head = f"contig_{i}_{len(seq)}bp"
                writer.write(new_head, seq)
                contig_dict[new_head] = head
                i += 1

    # Prokka command line
    meta = ['', '--metagenome '][metagenome]
    evalue = format(evalue, 'f')
    if proteins:
        proteins = f"--proteins {proteins} "
    if log is not None:
        log = f" 2> {log}"
    else:
        log = f" 2> {outdir}.log"

    cmd = f"prokka --outdir {outdir} --prefix {outdir} --locustag {locus_tag} \
--kingdom {kingdom} --evalue {evalue} --cpus {threads} {proteins}{meta}{temp_fa}{log}"

    call(cmd)
    os.remove(temp_fa)

    # In the outdir folder, write a tsv file to index shortened contig headers to the original headers
    with open(f"{outdir}/contig_index.tsv", 'w') as fh:
        for key, val in contig_dict.items():
            fh.write(f"{key}\t{val}\n")

    # Copy the genbank file out from the output directory
    call(f"cp {outdir}/{outdir}.gbk {outdir}.gb")

    # Rename the contig header back to the original headers using <contig_dict>
    _rename_contig_id(f"{outdir}.gb", contig_dict)

    # Remove files in the output dir if keep == False
    if not keep:
        call(f"rm -r {outdir}")


def trim_galore(
        fq1: str,
        fq2: str,
        quality: int = 20,
        gzip: bool = True,
        length: int = 20,
        log: str = 'trim_galore.log'):
    """
    Wrapper function of "trim_galore"

    Use the default settings of "trim_galore",
        e.g. Illumina P5 and P7 adapter sequences are used to trim reads

    Args:
        fq1: path-like
            The read-1 fastq file

        fq2: path-like
            The read-2 fastq file

        quality:
            phred33 score

        gzip:
            Compress the output fastq files or not

        length:
            The minimal length (bp) of reads to be retained

        log: path-like
            Append the stderr of trim_galore to the <log> file
    """
    gzip = ['', '--gzip '][gzip]
    call(f'trim_galore --paired --quality {quality} --phred33 --fastqc --illumina \
{gzip}--length {length} --max_n 0 --trim-n --retain_unpaired {fq1} {fq2} 2>> {log}')


def orf_finder(
        fasta: str,
        output: str,
        min_length: int = 75):
    """
    Use the command line tool NCBI ORFfinder to find Open Reading Frames
      in the input fasta file, and output a GTF file

    Args:
        fasta: path-like

        output: path-like
            The output GTF file

        min_length:
            Minimal length of ORF. 75 is the default of ORFfinder
    """
    # ORFfinder can only take fasta headers <= 50 characters
    # Create a temporary fasta file with short headers
    #   and a dictionary matching the short and the original headers
    temp_fa = _temp('temp', '.fa')  # for example, temp000.fa
    contig_dict = {}
    with FastaParser(fasta) as parser:
        with FastaWriter(temp_fa) as writer:
            for i, contig in enumerate(parser):
                head, seq = contig
                writer.write(str(i), seq)
                contig_dict[str(i)] = head

    # -outfmt 3: output format = feature table
    # -ml [int]: minimal length
    temp_table = _temp('temp', '.table')  # for example, temp000.table
    call(f"ORFfinder -in {temp_fa} -out {temp_table} -outfmt 3 -ml {min_length}")

    os.remove(temp_fa)

    with GtfWriter(output) as writer:
        with open(temp_table) as fh:
            while True:
                line1 = fh.readline().rstrip()
                line2 = fh.readline().rstrip()
                line3 = fh.readline().rstrip()
                if line1 == '':
                    break

                # line1: Get contig_number and orf_id
                contig_number = line1.split('_')[1].split(':')[0]
                orf_id = line1.split('>Feature lcl|')[1].split('_')[0]

                # Get the original contig header in the input fasta file
                header = contig_dict[contig_number]

                # line2: Get start and end (zero-based)
                start, end = line2.split('\t')[0:2]
                start, end = int(start), int(end)

                if start < end:
                    strand = '+'
                else:
                    strand = '-'
                    start, end = end, start

                # The start and end of GTF is 1-based
                feature = \
                    (header              ,  # 1 seqname   str
                     '.'                 ,  # 2 source    str
                     'CDS'               ,  # 3 feature   str
                     start + 1           ,  # 4 start     int
                     end + 1             ,  # 5 end       int
                     '.'                 ,  # 6 score     float
                     strand              ,  # 7 strand    str ('+', '-')
                     0                   ,  # 8 frame     int (0, 1, 2)
                     f"name \"{orf_id}\"")  # 9 attribute str
                writer.write(feature)
        os.remove(temp_table)



import unittest
import numpy as np
from ngslite.arrfunc import replace_zero_with


class TestArrfunc(unittest.TestCase):

    def test_replace_zero_with(self):
        arr1 = replace_zero_with(
            np.array([0, 1, 0, 2, 0, 3, 0]), 100
        )
        arr2 = np.array([100, 1, 100, 2, 100, 3, 100])
        diff = arr1 - arr2
        self.assertEqual(np.sum(diff), 0)


if __name__ == '__main__':
    unittest.main()



import os
from Bio import Phylo
from typing import Optional
from matplotlib.pyplot import Axes
from .fasta import FastaParser
from .lowlevel import call


def muscle(
        fasta: str,
        output: str,
        maxiters: int = 16,
        outformat: str = 'fasta',
        log: Optional[str] = None):
    """
    Wrapper function for MSUCLE sequence alignment

    Args:
        fasta: path-like
            The input fasta file

        output: path-like
            The output file

        maxiters:
            Maximum number of iterations (default 16)

        outformat:
            'fasta' (default)
            'html'
            'msf' (GCF MSF)
            'clw' (ClustalW)

        log: path-like
            The log file, default None -> <output>.log
    """
    outformat = {'fasta': '', 'html': '-html', 'msf': '-msf', 'clw': '-clw'}[outformat]

    if log is None: log = output + '.log'
    call(f'muscle -in {fasta} -out {output} -maxiters {maxiters} -log {log} {outformat}')


def simple_phylogeny(
        fasta: str,
        output: str,
        clustering: str = 'NJ',
        log: Optional[str] = None):
    """
    Make a tree file (newick/phylip format) from the aligned <fasta>
        using the command "clustalw -tree"

    Args:
        fasta: path-like
            The input fasta file, sequences must be aligned

        output: path-like
            The output tree file

        clustering:
            clustering method, 'NJ' or 'UPGMA'

        log: path-like
            The log file, default None -> <output>.log
    """
    with FastaParser(fasta) as parser:
        head, seq = parser.next()
        if set(seq.upper()) == set('ACGT-'): type_ = 'DNA'
        else: type_ = 'PROTEIN'

    if log is None: log = output + '.log'
    call(f'clustalw -tree -infile={fasta} -type={type_} -CLUSTERING={clustering} > {log}')

    # Rename the .ph file automatically generated by clustalw
    ph_file = fasta[:fasta.rfind('.')] + '.ph'
    os.rename(ph_file, output)


def draw_tree(file: str, ax: Axes):
    """
    Wrapper function for Bio.Phylo.read() and Bio.Phylo.draw()
        to draw a phylogenetic tree from the input tree <file>

    Args:
        file: path-like
            The input tree file in newick/phylip format

        ax: matplotlib Axes object
            The ax object in which the tree will be drawn
    """
    tree = Phylo.read(file, 'newick')
    Phylo.draw(tree, axes=ax)



import random
import subprocess
import numpy as np
from .fasta import read_fasta
from .dnatools import rev_comp
from .lowlevel import call, _temp, gzip, printf
from .file_conversion import fq_to_fa


def jellyfish_count(file, k, output, min_count=1, hash_size='100M', threads=4, canonical=True):
    """
    Wrapper function for the jellyfish k-mer count commands

    From a high-level it takes an input fastq or fastq <file>, and eventually
        exports an <output> fasta file of k-mer counts

    All the intermediate files are deleted

    Args:
        file: str, path-like
            The input fasta or fastq file, also accepts .gz compressed format

        k: int
            The size of the k-mer

        output: str, path-like
            The output fasta file, which cannot be the same as the input <file>

        min_count: uint
            K-mers with counts < min_count will be excluded

        hash_size: str,
            e.g. '100M' for 100 million; '1G' for 1 billion

        threads: int (uint32)

        canonical: bool
            If True, 'AAAC' will be the same as 'GTTT'
    """
    # The <output> fasta should NOT be identical to the input fasta <file>
    if output == file: return

    input_name = file  # The original input name

    # Unzip
    if file.endswith('.gz'):
        gzip(file, keep=True)
        file = file[:-3]

    # Convert fastq to fasta
    if file.endswith('.fastq') or file.endswith('.fq'):
        fq_to_fa(file, keep=True)
        # If the input <file> is not equal to the original input name, then don't keep it
        if not file == input_name: call(f"rm {file}")

    if file.endswith('.fastq'): file = file[:-6] + '.fa'
    elif file.endswith('.fq'): file = file[:-3] + '.fa'

    # Count k-mers in the input fastq or fasta file
    canonical = ['', '-C '][canonical]
    temp_jf = _temp('temp', '.jf')  # for example, temp000.jf
    call(f"jellyfish count -m {k} -s {hash_size} -t {threads} -o {temp_jf} {canonical}{file}")

    # Remove the counted file if it's not the original input file
    if not file == input_name: call(f"rm {file}")

    # Dump the .jf file into a .fa file
    call(f"jellyfish dump -L {min_count} --output {output} {temp_jf}")

    # The .jf file is not needed, so remove it.
    call(f"rm {temp_jf}")


def read_kmers(jf_fa, min_count=1):
    """
    Args:
        jf_fa: str, path-like
            The fasta exported by jellyfish

        min_count: int
            Kmers >= min_count will be included

    Returns:
        two arrays:
            kmers: list of str
                Each str is the sequence of a kmer

            counts: np.ndarray, dtype=np.uint32
                Count of kmers in the jf_fa file
    """
    kmers = []
    counts = []
    with open(jf_fa, 'r') as fh:
        while True:
            line1 = fh.readline().strip() # header = count
            line2 = fh.readline().strip() # sequence
            if line1 == '' or line2 == '':
                break
            c = int(line1[1:])
            if c >= min_count:
                kmers.append(line2)
                counts.append(c)
    return kmers, np.array(counts, dtype=np.uint32)


def kmer_2D_points(jf_fa_1, jf_fa_2, min_count=1):
    """
    Args:
        jf_fa_1: str, path-like
            The fasta exported by jellyfish

        jf_fa_2: str, path-like
            The fasta exported by jellyfish

        min_count: int
            Kmers >= min_count will be included from the jellyfish fasta

    Returns:
        three arrays:
            kmers: list of str
                Each str is the sequence of a kmer

            counts_1: np.ndarray, dtype=np.uint32
                Count of kmers in the jf_fa_1 file

            counts_2: np.ndarray, dtype=np.uint32
                Count of kmers in the jf_fa_2 file
    """
    # --- Read the jf_fa_1 file into a dictionary --- #

    # D1 = {'AAAA': count,
    #       'AAAC': count}
    D1 = {}
    with open(jf_fa_1, 'r') as fh:
        while True:
            line1 = fh.readline().strip()
            line2 = fh.readline().strip()
            if line1 == '' or line2 == '':
                break
            c = int(line1[1:])
            if c >= min_count:
                D1.setdefault(line2, c)

    # --- Read the jf_fa_2 file into a dictionary --- #

    # D2 = {'AAAA': count,
    #       'AAAC': count}
    D2 = {}
    with open(jf_fa_2, 'r') as fh:
        while True:
            line1 = fh.readline().strip()
            line2 = fh.readline().strip()
            if line1 == '' or line2 == '':
                break
            c = int(line1[1:])
            if c >= min_count:
                D2.setdefault(line2, c)

    # --- Combine the two dictionaries into one --- #

    # D3 = {'AAAA': (count_1, count_2),
    #       'AAAC': (count_1, count_2)}
    D3 = {}
    for kmer, c1 in D1.items():
        c2 = D2.get(kmer, 0)
        D3.setdefault(kmer, (c1, c2))
        # If the kmer exit in D2, then pop it out
        if c2 != 0:
            D2.pop(kmer)

    # Anything left in D2 is NOT present in D1
    for kmer, c2 in D2.items():
        D3.setdefault(kmer, (0, c2))

    kmers, counts_1_2 = zip(*D3.items())
    counts_1, counts_2 = zip(*counts_1_2)

    return kmers, np.array(counts_1, dtype=np.uint32), np.array(counts_2, dtype=np.uint32)


def save_kmer_2D_points(kmers, x, y, file, x_name='x', y_name='y', description=''):
    """
    This method writes a tab-delimited file to store 2D points of kmers.

    The first line is the description text.

    The second line is the header line:
        'k-mer' \t <x_name> \t <y_name>

    From the third line, each line is a kmer:
        <kmer_seq> \t <x_coordinate> \t <y_coordinate>

    Here's an example:
        k-mer  \t <x_name> \t <y_name>
        AAAAAA \t 10       \t 10
        AAAAAC \t 1        \t 1

    Args:
        kmers: list of str
            Each str is the sequence of a kmer

        x: list of num, or numpy array, dtype = np.int or np.float
            Each num is the x coordinate of the kmer in the <kmers> list

        y: list of num, or numpy array, dtype = np.int or np.float
            Each num is the y coordinate of the kmer in the <kmers> list

        file: str, path-like
            The output tab-delimited file

        x_name: str
            The column name for the x coordinates.
            Usually it's the source filename from which k-mers were counted

        y_name: str
            The column name for the y coordinates.
            Usually it's the source filename from which k-mers were counted

        description: str
            Extra description of this file, which will be written in the first line

    Returns: None
    """
    if len(kmers) != len(x):
        printf("The number of 'kmers' is not equal to that of 'x'")
        return
    if len(kmers) != len(y):
        printf("The number of 'kmers' is not equal to that of 'y'")
        return

    if not file.endswith('.tsv'):
        file = file + '.tsv'

    # Replace any '\n' in the <description> with '   '
    #   since <description> should be limited to the first line
    if '\n' in description:
        description = '   '.join(description.split('\n'))

    with open(file, 'w') as fh:
        fh.write(description + '\n')
        fh.write('k-mer\t{}\t{}\n'.format(x_name, y_name))
        for i in range(len(kmers)):
            fh.write('{}\t{}\t{}\n'.format(kmers[i], x[i], y[i]))


def read_kmer_2D_points(file, dtype=np.uint32, fraction=1.):
    """
    This function reads the tab-delimited file written by the method <save_kmer_2D_points>.

    The first line is the comment/description text.

    The second line is the header line:
        'k-mer' \t <x_name> \t <y_name>

    From the third line, each line is a kmer:
        <kmer_seq> \t <x_coordinate> \t <y_coordinate>

    Here's an example:
        k-mer  \t <x_name> \t <y_name>
        AAAAAA \t 10       \t 10
        AAAAAC \t 1        \t 1

    Args:
        file: str, path-like
            The input tsv file written by the method <save_kmer_2D_points>

        dtype: numpy data types
            Number data types for x and y coordinates; np.uint32 is the default

        fraction: float
            The fraction of k-mers to be randomly retrieved from the file
            Default = 1., i.e. all k-mers

    Returns:
        three arrays:
            kmers: list of str
                Each str is the sequence of a kmer

            x: np.ndarray, dtype=np.uint32
                x coordinate of the kmer in the <kmers> list

            y: np.ndarray, dtype=np.uint32
                y coordinate of the kmer in the <kmers> list

        three strings:
            x_name: str
                The column name for the x coordinates.
                Usually it's the source filename from which k-mers were counted.

            y_name: str
                The column name for the y coordinates.
                Usually it's the source filename from which k-mers were counted.

            description: str
                Extra description of this file
    """
    data = []

    with open(file, 'r') as fh:
        description = fh.readline().rstrip()
        _, x_name, y_name = fh.readline().rstrip().split('\t')
        for line in fh:
            if random.uniform(0, 1) <= fraction:
                data.append(
                    tuple(
                        line.rstrip().split('\t') # 'kmer\tx\ty' => (kmer, x, y)
                    )
                )

    kmers, x, y = zip(*data) # Unzip into three lists

    return kmers, np.array(x, dtype=dtype), np.array(y, dtype=dtype), x_name, y_name, description


def filter_kmer_2D_points(kmers, x, y, ref_fasta=None, min_count=None, func=None):
    """
    Retrieve kmers that fits the following criteria:
        1) Appearing in the reference genome <ref_fasta>. If ref genome is None, then no filtering.
        2) Greater or equal than min_count. If min_count is None, then no filtering.
        3) Passing the criteria of the function <func>. If func is None, then no filtering.

    Args:
        kmers: list of str
            Each str is the sequence of a kmer

        x: np.ndarray, dtype = int or float
            x coordinate of the kmer in the <kmers> list

        y: np.ndarray, dtype = int or float
            y coordinate of the kmer in the <kmers> list

        ref_fasta: str
            If str, then it's the fasta file path of the reference genome
            If None, then don't filter

        min_count: single number or a tuple of (number, number), where the number could be int or float
            If a single number, then filter both x and y with that number
            If (number_x, number_y), then filter x and y with number_x and number_y, respectively
            If None, then don't filter

        func: function object,
            This function should have three input parameters, for <kmers>, <x> and <y>, respectively;
                and should return True/False
            If None, then don't filter

    Returns:
        three arrays:
            [1] list of str
                Each str is the sequence of a kmer

            [2] np.ndarray, dtype determined by the input
                x coordinate of the kmer in the <kmers> list

            [3] np.ndarray, dtype determined by the input
                y coordinate of the kmer in the <kmers> list
    """
    # From a high level, the function walks through the input kmer list,
    #     for each kmer see if it fits the criteria
    #     and eventually return the k-mers fitting the criteria

    kmers_new, x_new, y_new = kmers, x, y

    # --- Filter by the reference genome sequence --- #
    if not ref_fasta is None:

        # Push the new ones into old ones and empty the new ones
        kmers_old, x_old, y_old = kmers_new, x_new, y_new
        kmers_new, x_new, y_new = [], [], []

        ref_seq = read_fasta(ref_fasta)[0][1]

        # Convert to UPPER CASE!
        ref_seq = ref_seq.upper()

        # Build a dictionary of k-mers of the reference sequence: {'AAAA': True, 'AAAC': True}
        # A particular k-mer may appear in the reference sequence more than once,
        #     but it would be unique as a key in the dictionary
        ref_dict = {}

        k = len(kmers[0])
        # For each k-mer in the forward sequence of the reference
        for i in range(len(ref_seq) - k + 1):
            ref_dict.setdefault(ref_seq[i:i+k], True)

        # Reverse complementary of the reference genome sequence
        rc_ref_seq = rev_comp(ref_seq)

        # For each k-mer in the reverse complementary sequence of the reference
        for i in range(len(rc_ref_seq) - k + 1):
            ref_dict.setdefault(rc_ref_seq[i:i+k], True)

        # Now the k-mer dictionary of the reference sequence has been built

        # For each input k-mer, see if it appears in the reference genome
        # If yes
        for i, kmr in enumerate(kmers_old):
            if ref_dict.get(kmr, None):
                kmers_new.append(kmr)
                x_new.append(x_old[i])
                y_new.append(y_old[i])

    # --- Filter by minimum count --- #
    if not min_count is None:

        # Push the new ones into old ones and empty the new ones
        kmers_old, x_old, y_old = kmers_new, x_new, y_new
        kmers_new, x_new, y_new = [], [], []

        if isinstance(min_count, int):
            x_min, y_min = min_count, min_count
        if isinstance(min_count, tuple) or isinstance(min_count, list):
            x_min, y_min = min_count

        for i, kmr in enumerate(kmers):
            if x_old[i] >= x_min and y_old[i] >= y_min:
                kmers_new.append(kmr)
                x_new.append(x_old[i])
                y_new.append(y_old[i])

    # --- Filter by the function object --- #
    if not func is None:

        # Push the new ones into old ones and empty the new ones
        kmers_old, x_old, y_old = kmers_new, x_new, y_new
        kmers_new, x_new, y_new = [], [], []

        for i, kmr in enumerate(kmers):
            if func(kmr, x_old[i], y_old[i]):
                kmers_new.append(kmr)
                x_new.append(x_old[i])
                y_new.append(y_old[i])

    return kmers_new, np.array(x_new), np.array(y_new)


def sample_kmer_2D_points(kmers, x, y, fraction):
    """
    Randomly sample kmers

    Args:
        kmers: list of str
            Each str is the sequence of a kmer

        x: np.ndarray, dtype = int or float
            x coordinate of the kmer in the <kmers> list

        y: np.ndarray, dtype = int or float
            y coordinate of the kmer in the <kmers> list

        fraction: float (0..1)
            The fraction of kmers to be sampled

    Returns:
        three arrays:
            [1] list of str
                Each str is the sequence of a kmer

            [2] np.ndarray, dtype determined by the input
                x coordinate of the kmer in the <kmers> list

            [3] np.ndarray, dtype determined by the input
                y coordinate of the kmer in the <kmers> list
    """
    ret_kmer = []
    ret_x = []
    ret_y = []
    for i, kmr in enumerate(kmers):
        if random.uniform(0, 1) <= fraction:
            ret_kmer.append(kmr)
            ret_x.append(x[i])
            ret_y.append(y[i])
    return ret_kmer, np.array(ret_x), np.array(ret_y)


def build_kmer_dict(kmers, x, y):
    """
    Args:
        kmers: list of str
            Each str is the sequence of a kmer

        x: list of num, or numpy array, dtype = np.int or np.float
            Each num is the x coordinate of the kmer in the <kmers> list

        y: list of num, or numpy array, dtype = np.int or np.float
            Each num is the y coordinate of the kmer in the <kmers> list

    Returns: dict
        {'AAAA': (x_1, y_1),
         'AAAC': (x_2, y_2)}
        where the coordinates x_1, y_1, x_2, y_2 could be int or float
        depending on the input <x> and <y>
    """
    return {kmr: xy for kmr, xy in zip(kmers, zip(x, y))}


def get_kmer_2D_points(seq, kmer_dict):
    """
    Convert the input <seq> into k-mers and
        find the 2D positions of those k-mers in the <kmer_dict>

    Args:
        seq: str
            DNA sequence

        kmer_dict: dict
            {'AAAA': (x_1, y_1),
             'AAAC': (x_2, y_2)}
            where the coordinates x_1, y_1, x_2, y_2 could be int or float

    Returns:
        three arrays:
            kmers: list of str
                Each str is the sequence of a kmer

            x: np.ndarray, dtype determined by the input kmer_dict
                x coordinate of the kmer in the <kmers> list

            y: np.ndarray, dtype determined by the input kmer_dict
                y coordinate of the kmer in the <kmers> list
    """
    # Overall, the function does:
    #   1) Walk through each k-mer in the <seq>
    #   2) If the k-mer appears in the <kmer_dict>, put it in the new dictionary
    #      A particular k-mer could appear more than once in the <seq>,
    #          but it would appear only once in the new dictionary
    #   3) Convert the dictionary of k-mers into three lists: kmers, x, y

    # The length of each k-mer
    k = len(tuple(kmer_dict.keys())[0])

    # Convert to UPPER CASE!
    seq = seq.upper()

    # Build a k-mer dictionary for the input <seq>:
    #     {'AAAA': (x, y), 'AAAC': (x, y), ...}
    # A particular k-mer may appear in the input <seq> more than once,
    #     but it would be unique as a key in the dictionary
    seq_kmer_dict = {}

    # For each k-mer in the input sequence,
    #     see if the k-mer is available in the kmer_dict.
    # If available, then get the counts of the k-mer from kmer_dict
    for i in range(len(seq) - k + 1):

        kmr = seq[i:(i+k)]
        rc_kmr = rev_comp(kmr)

        # Either kmr or rc_kmr should be found in the kmer_dict
        #     because of canonical k-mer counting
        # If kmr is found in kmer_dict,
        #     then there is no need to try to find rc_kmr in the kmer_dict
        if kmer_dict.get(kmr, None):
            seq_kmer_dict.setdefault(kmr, kmer_dict[kmr])

        elif kmer_dict.get(rc_kmr, None):
            seq_kmer_dict.setdefault(rc_kmr, kmer_dict[rc_kmr])

    if seq_kmer_dict:
        kmers, xy = zip(*seq_kmer_dict.items())
        x, y = zip(*xy)
    else:
        kmers, x, y = [], [], []

    return kmers, np.array(x), np.array(y)


def kmers_base_content(kmers, base):
    """
    Args:
        kmers: list of str
            An array of k-mers

        base: str
            e.g., 'A' gives the A content; 'GC' gives the GC content; 'ACGT' should give 1.0

    Returns:
        np.ndarray, dtype=np.float32:
            An array of base contents of the input k-mers
    """
    ret = np.zeros((len(kmers), ), dtype=np.float32)
    for b in set(base):
        ret = ret + np.array([kmr.count(b) / len(kmr) for kmr in kmers], dtype=np.float32)
    return ret


def fastq_to_saved_kmer_2D_points(fq1, fq2, k, output, min_count=1, hash_size='100M', threads=4, description=''):
    """
    The full implementation of the pipeline:
        (1) Input two fastq files
        (2) Count k-mers in each fastq file
        (3) Merge (outer join) the k-mer counts of the two files
        (4) Write a tab-separated file (.tsv) of k-mer 2D points

    Args:
        fq1: str, path-like
            The first fastq file

        fq2: str, path-like
            The second fastq file

        k: uint
            The size of the k-mer in bp

        output: str, path-like
            The output tsv file

        min_count: int
            K-mers with count < min_count will be set to zero

        hash_size: str
            The option for the jellyfish_count() function; '100M' = 100 million, '1G' = 1 billion

        threads: uint
            The number of CPU to be used for the jellyfish_count() function

        description: str
            The description to be written into the output tsv file
    """
    # fastq -> fasta reported by jellyfish
    jellyfish_count(file=fq1, k=k, output='1.fa', min_count=min_count, hash_size=hash_size, threads=threads)
    jellyfish_count(file=fq2, k=k, output='2.fa', min_count=min_count, hash_size=hash_size, threads=threads)

    # Merge two fasta reported by jellyfish -> k-mer 2D points
    kmers, counts_1, counts_2 = kmer_2D_points(jf_fa_1='1.fa', jf_fa_2='2.fa', min_count=min_count)

    # Save k-mer 2D points as the <output> tsv file
    save_kmer_2D_points(kmers=kmers, x=counts_1, y=counts_2, file=output, x_name=fq1, y_name=fq2, description=description)

    # Remove the two fasta files exported by jellyfish
    subprocess.check_call('rm 1.fa 2.fa', shell=True)



from setuptools import setup
import ngslite


setup(
    name='ngslite',
    version=ngslite.__version__,
    description='Light-weight functions for next-generation sequencing (NGS) data analysis',
    url='https://github.com/linyc74/ngslite',
    author='Yu-Cheng Lin',
    author_email='yclin.python@gmail.com',
    license='MIT',
    packages=['ngslite'],
    python_requires='>3.6',
    install_requires=['numpy', 'pandas', 'biopython', 'scipy', 'dna_features_viewer'],
    zip_safe=False
)


# In command line type in: "python setup.py sdist"



import unittest
from ngslite.dataclass import GenericFeature
from ngslite.genbank_write import wrap_line_by_word, wrap_line_by_char, \
    generic_feature_to_genbank_text, translate_feature


class TestGenbankWrite(unittest.TestCase):

    def test_wrap_line_by_word(self):
        line = '''This is a good world with good intentions, where people love each other every day.'''
        wrapped = '''\
        This is a good world with
        good intentions, where
        people love each other
        every day.'''
        self.assertEqual(
            wrapped,
            wrap_line_by_word(
                line=line, length=33, indent=8, sep=' ', keep_sep=False)
        )

        line = '''This-is-a-good-world-with-good-intentions,-where-people-love-each-other-every-day.'''
        wrapped = '''\
        This-is-a-good-world-
        with-good-intentions,-
        where-people-love-each-
        other-every-day.'''
        self.assertEqual(
            wrapped,
            wrap_line_by_word(
                line=line, length=33, indent=8, sep='-', keep_sep=True)
        )

    def test_wrap_line_by_char(self):
        line = 'ABCDEFGHIJKLMNOPQRSEUVWXYZ'
        wrapped = '''\
        ABCDEF
        GHIJKL
        MNOPQR
        SEUVWX
        YZ'''
        self.assertEqual(
            wrapped,
            wrap_line_by_char(line=line, length=14, indent=8)
        )

    def test_generic_feature_to_genbank_text(self):
        feature = GenericFeature(
            seqname='yuchenglin',
            type_='DNA',
            start=74,
            end=1983,
            strand='-',
            attributes=[
                ('gene', 'engineering'),
                ('accession', 'R97b47402'),
                ('age', 36),
                ('distance', 42.195),
                ('description', 'Yu-Cheng is a computational biologist who tries to do something \
cool at the intersection of biology and programming. He also likes to run and play guitar.'),
                ('translation', 'GIBBERISHGIBBERISHGIBBERISHGIBBERISHGIBBERISHGIBBERISHGIBBERISH'),
            ],
            regions=[(74, 740, '-'), (1000, 1100, '-'), (1200, 1300, '-'), (1400, 1500, '-'), (1900, 1983, '-')],
            frame=1,
            partial_start=True,
            partial_end=True
        )

        text = '''\
     DNA             complement(join(<74..740,1000..1100,1200..1300,1400..1500,
                     1900..>1983))
                     /gene="engineering"
                     /accession="R97b47402"
                     /age=36
                     /distance=42.195
                     /description="Yu-Cheng is a computational biologist who
                     tries to do something cool at the intersection of biology
                     and programming. He also likes to run and play guitar."
                     /translation="GIBBERISHGIBBERISHGIBBERISHGIBBERISHGIBBERIS
                     HGIBBERISHGIBBERISH"
'''
        self.assertEqual(
            text,
            generic_feature_to_genbank_text(feature=feature)
        )

    def test_translate_feature(self):
        feature = GenericFeature(
            seqname='seqname',
            type_='CDS',
            start=1,
            end=250,
            strand='-',
            attributes=[('codon_start', 1)],
            regions=[(1, 50, '-'), (101, 150, '-'), (201, 250, '-')],
            partial_start=False,
            partial_end=False
        )

        sequence = '''\
ttcgtcaccggtatcgctttccactttgcctaccgcaccatggacaagca\
aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa\
ggaagacgctctcaacctgctgcccactggtcaccttggcactgaggagc\
aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa\
aggctgccgacattgagaggcgccagagcgttactgccgagaagcactaa'''

        translation = '''\
MVLLGSNALAPLNVGSLLLSAKVTSGQQVESVFLLVHGAVGKVESDTGDE'''

        self.assertEqual(
            translation,
            translate_feature(feature=feature, sequence=sequence)
        )


if __name__ == '__main__':
    unittest.main()




from .lowlevel import call


def sam_to_bam(file: str, keep: bool = True):
    """
    Wrapper function of "samtools view" to convert sam into bam

    Args:
        file: path-like

        keep:
            Keep the input file or not
    """
    # -S: input is a sam
    # -b: output is a bam
    # -h: include header section
    call(f"samtools view -S -b -h {file} > {file[:-4]}.bam")
    if not keep:
        call(f"rm {file}")


def bam_to_sam(file: str, keep: bool = True):
    """
    Wrapper function of "samtools view" to convert bam into sam

    Args:
        file: path-like

        keep:
            Keep the input file or not
    """
    # -h: include header section
    call(f"samtools view -h {file} > {file[:-4]}.sam")
    if not keep:
        call(f"rm {file}")


def fq_to_fa(file: str, keep: bool = True):
    """
    Wrapper function of "seqtk" to convert fastq into fasta

    Args:
        file: path-like

        keep:
            Keep the input file or not
    """
    if file.endswith('.fastq'):
        output = file[:-6] + '.fa'
    elif file.endswith('.fq'):
        output = file[:-3] + '.fa'
    else:
        output = file + '.fa'

    call(f"seqtk seq -A {file} > {output}")
    if not keep:
        call(f"rm {file}")


def vcf_to_bcf(file: str, keep: bool = True):
    """
    Wrapper function of "bcftools view" to convert VCF into BCF

    Args:
        file: path-like

        keep:
            Keep the input file or not
    """
    # -Ou: output uncompressed bcf
    # -o <file_out>
    call(f"bcftools view -Ou -o {file[:-4]}.bcf {file}")
    if not keep:
        call(f"rm {file}")


def bcf_to_vcf(file: str, keep: bool = True):
    """
    Wrapper function of "bcftools view" to convert BCF into VCF

    Args:
        file: path-like

        keep:
            Keep the input file or not
    """
    # -Ov: output uncompressed vcf
    # -o <file_out>
    call(f"bcftools view -Ov -o {file[:-4]}.vcf {file}")
    if not keep:
        call(f"rm {file}")



import unittest
from ngslite.dataclass import GenericFeature, FeatureArray


class TestGenericFeature(unittest.TestCase):

    def setUp(self):
        self.feature = GenericFeature(
            seqname='seqname', type_='DNA',
            start=1, end=1000, strand='+',
            attributes=[('key_1', 'val_1'), ('key_2', 'val_2.0'), ('key_2', 'val_2.1')],
            tags=None,
            regions=[(1, 100, '+'), (200, 300, '+'), (900, 1000, '+')],
            frame=2, partial_start=True, partial_end=False
        )

    def test___len__(self):
        self.assertEqual(len(self.feature), 1000)

    def test_get_attribute(self):
        self.assertEqual(self.feature.get_attribute('key_1'), 'val_1')
        self.assertEqual(self.feature.get_attribute('key_2'), ['val_2.0', 'val_2.1'])

    def test_set_attribute(self):
        self.feature.set_attribute(key='key_1', val='new_val_1')
        self.feature.set_attribute(key='key_2', val='new_val_2')
        self.assertListEqual(
            self.feature.attributes,
            [('key_1', 'new_val_1'), ('key_2', 'new_val_2')]
        )

    def test_add_attribute(self):
        self.feature.add_attribute(key='key_3', val='val_3')
        self.assertListEqual(
            self.feature.attributes,
            [('key_1', 'val_1'), ('key_2', 'val_2.0'), ('key_2', 'val_2.1'), ('key_3', 'val_3')]
        )

    def test_remove_attribute(self):
        self.feature.remove_attribute(key='key_2')
        self.assertEqual(self.feature.attributes, [('key_1', 'val_1')])
        self.feature.remove_attribute(key='key_1')
        self.assertEqual(self.feature.attributes, [])


class TestFeatureArray(unittest.TestCase):

    def setUp(self):
        self.feature_list = []
        for i in range(5):
            f = GenericFeature(
                seqname='seqname', type_='CDS',
                start=i*20+1, end=i*20+10, strand='+',
            )
            self.feature_list.append(f)

        self.feature_array = FeatureArray(
            seqname='seqname', genome_size=100,
            features=self.feature_list, circular=True
        )

    def test___getitem__(self):
        self.assertEqual(
            id(self.feature_array[0]),
            id(self.feature_list[0])
        )

        self.assertNotEqual(
            id(self.feature_array),
            id(self.feature_array[:])
        )

        for id1, id2, in zip(self.feature_array[:], self.feature_array):
            self.assertEqual(id1, id2)

    def test___add__(self):
        new_feature = GenericFeature(
            seqname='seqname', type_='CDS',
            start=11, end=15, strand='+',
        )

        new_array = FeatureArray(
            seqname='seqname', genome_size=100,
            features=[new_feature], circular=True
        )

        new_array = self.feature_array + new_array

        self.assertNotEqual(id(new_array), id(self.feature_array))
        self.assertEqual(id(new_array[0]), id(self.feature_array[0]))
        self.assertEqual(id(new_array[2]), id(self.feature_array[1]))
        self.assertEqual(id(new_array[3]), id(self.feature_array[2]))
        self.assertEqual(id(new_array[4]), id(self.feature_array[3]))
        self.assertEqual(id(new_array[5]), id(self.feature_array[4]))

    def test_sort(self):
        pass

    def test_append(self):
        pass

    def test_subset(self):
        pass

    def test_offset(self):
        pass

    def test_crop(self):
        pass

    def test_reverse(self):
        pass

    def test_pop(self):
        pass


if __name__ == '__main__':
    unittest.main()




from typing import Dict, List, Union
from .fasta import FastaParser
from .gtftools import read_gtf
from .dataclass import Chromosome, FeatureArray


def read_fasta_gtf(
        fasta: str,
        gtf: str,
        as_dict: bool = False,
        circular: bool = False) \
        -> Union[List[Chromosome], Dict[str, Chromosome]]:
    """
    Read fasta and gtf simultaneously into Chromosome objects, i.e. annotated genomes

    Args:
        fasta: path-like

        gtf: path-like

        as_dict:
            Return as dictionary or not

        circular:
            Shape of the DNA molecule

    Returns: list of Chromosome, or dict

        [Chromosome_1, Chromosome_2, ...]

        or

        {
            Chromosome_1.seqname: Chromosome_1,
            Chromosome_2.seqname: Chromosome_2, ...
        }
    """
    chromosomes = []

    feature_dict = read_gtf(gtf, as_dict=True)
    # feature_dict has the data structure:
    # {
    #   seqname: [GtfFeature, ...], ...
    # }

    with FastaParser(fasta) as parser:
        for seqname, sequence in parser:

            gtf_features = feature_dict.get(seqname, [])

            features = FeatureArray(
                seqname=seqname,
                genome_size=len(sequence),
                features=gtf_features,
                circular=circular)

            c = Chromosome(
                seqname=seqname,
                sequence=sequence,
                features=features,
                circular=False)

            chromosomes.append(c)

    if as_dict:
        return {chrom.seqname: chrom for chrom in chromosomes}

    return chromosomes
